/*
 * Copyright (c) 2020 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-BSD-5-Clause-Nordic
 */

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/logging/log.h>
#include <zephyr/drivers/gpio.h>

#include <nrfx_gpiote.h>
#include "hal/nrf_gpio.h"

#include <nrfx_timer.h>


#include <helpers/nrfx_gppi.h>
#if defined(DPPI_PRESENT)
#include <nrfx_dppi.h>
#else
#include <nrfx_ppi.h>
#endif

LOG_MODULE_REGISTER(app);

// https://github.com/nrfconnect/sdk-zephyr/blob/master/samples/boards/nrf/nrfx/src/main.c
// https://devzone.nordicsemi.com/f/nordic-q-a/34547/ppi-configuration-to-count-pulses

#define INPUT_PIN 7
#define INPUT_PIN_1 8


#define COUNT_READ_INTERVAL 1000 // Interval in ms to read the counting timer


static const nrfx_timer_t m_timer_count = NRFX_TIMER_INSTANCE(0);

static const nrfx_timer_t m_timer_read  = NRFX_TIMER_INSTANCE(1);

nrfx_gpiote_t gpiote = NRFX_GPIOTE_INSTANCE(0);

//nrf_ppi_channel_t ppi_channel_1, ppi_channel_2;

void timer_handler_count(nrf_timer_event_t event_type, void * p_context)
{

}

void timer_handler_read(nrf_timer_event_t event_type, void * p_context)
{
    uint32_t count_0 = nrfx_timer_capture_get(&m_timer_count, NRF_TIMER_CC_CHANNEL0);
	uint32_t freq_0 = 0;
    if(count_0 > 0)
    {
	   freq_0 = (1000*count_0)/COUNT_READ_INTERVAL;
	   printk("Freq [0]: %d [Hz]\n",freq_0);
    }
	//printk("Count[0]: %d\n",count_0);

	//edited
	uint32_t count_1 = nrfx_timer_capture_get(&m_timer_count, NRF_TIMER_CC_CHANNEL1);
	uint32_t freq_1 = 0;
    if(count_1 > 0)
    {
	   freq_1 = (1000*count_1)/COUNT_READ_INTERVAL;
	   printk("Freq [1]: %d [Hz]\n",freq_1);
    }
	//printk("Count[1]: %d\n",count_1);
	//edited
	printk("--------------------------\n");
}




void gpio_handler(nrfx_gpiote_pin_t pin, nrf_gpiote_polarity_t action)
{
    printk("gpio_handler\n");
}


/**
 * @brief Function for configuring: INPUT_PIN pin for input sensing
 */
static void gpiote_init(void) {
    nrfx_err_t err;
    nrfx_gpiote_pin_t pin = NRF_GPIO_PIN_MAP(0, INPUT_PIN);
    nrfx_gpiote_pin_t pin_1 = NRF_GPIO_PIN_MAP(0, INPUT_PIN_1);

    // Allocate separate GPIOTE channels for each pin
    uint8_t channel_0, channel_1;

    err = nrfx_gpiote_channel_alloc(&gpiote, &channel_0);
    if (err != NRFX_SUCCESS) {
        LOG_ERR("gpiote: Error allocating channel for INPUT_PIN: %08x", err);
    } else {
        LOG_INF("gpiote: Channel %d allocated for INPUT_PIN", channel_0);
    }

    err = nrfx_gpiote_channel_alloc(&gpiote, &channel_1);
    if (err != NRFX_SUCCESS) {
        LOG_ERR("gpiote: Error allocating channel for INPUT_PIN_1: %08x", err);
    } else {
        LOG_INF("gpiote: Channel %d allocated for INPUT_PIN_1", channel_1);
    }

    // Configure separate trigger configurations for each pin
    nrfx_gpiote_trigger_config_t trigger_config_0 = {
        .trigger = NRFX_GPIOTE_TRIGGER_LOTOHI,
        .p_in_channel = &channel_0,
    };

    nrfx_gpiote_trigger_config_t trigger_config_1 = {
        .trigger = NRFX_GPIOTE_TRIGGER_LOTOHI,
        .p_in_channel = &channel_1,
    };

    // Configure separate input pin configurations for each pin
    nrfx_gpiote_input_pin_config_t input_pin_config_0 = {
        .p_trigger_config = &trigger_config_0,
    };

    nrfx_gpiote_input_pin_config_t input_pin_config_1 = {
        .p_trigger_config = &trigger_config_1,
    };

    // Configure GPIOTE inputs
    err = nrfx_gpiote_input_configure(&gpiote, pin, &input_pin_config_0);
    if (err != NRFX_SUCCESS) {
        LOG_ERR("gpiote: Error configuring INPUT_PIN: %08x", err);
    } else {
        LOG_INF("gpiote: INPUT_PIN configured");
    }

    err = nrfx_gpiote_input_configure(&gpiote, pin_1, &input_pin_config_1);
    if (err != NRFX_SUCCESS) {
        LOG_ERR("gpiote: Error configuring INPUT_PIN_1: %08x", err);
    } else {
        LOG_INF("gpiote: INPUT_PIN_1 configured");
    }

    // Enable GPIOTE triggers for both pins
    nrfx_gpiote_trigger_enable(&gpiote, pin, false);
    nrfx_gpiote_trigger_enable(&gpiote, pin_1, false);

    // Verify GPIOTE event addresses
    printk("GPIOTE Event for INPUT_PIN: 0x%08x\n",
           nrf_gpiote_event_address_get(NRF_GPIOTE, nrfx_gpiote_in_event_get(&gpiote, INPUT_PIN)));

    printk("GPIOTE Event for INPUT_PIN_1: 0x%08x\n",
           nrf_gpiote_event_address_get(NRF_GPIOTE, nrfx_gpiote_in_event_get(&gpiote, INPUT_PIN_1)));
}


static void timer_init(void)
{
	// Configure m_timer_count for counting of low to high events on GPIO
	nrfx_err_t          err;


    uint32_t base_frequency = NRF_TIMER_BASE_FREQUENCY_GET(m_timer_count.p_reg);  // base freq 16Mhz


	nrfx_timer_config_t timer_cfg = {
		.mode      = NRF_TIMER_MODE_LOW_POWER_COUNTER,
		.bit_width = NRF_TIMER_BIT_WIDTH_32,
		.p_context = NULL,
        .interrupt_priority = IRQ_PRIO_LOWEST,
        .frequency = base_frequency,
	};

	err = nrfx_timer_init(&m_timer_count, &timer_cfg, timer_handler_count);
	if (err != NRFX_SUCCESS) {
		printk("nrfx_timer_init failed with: %d\n", err);
	}
	// Configure m_timer_read for reading the counter timer at a given interval COUNT_READ_INTERVAL
    timer_cfg.mode = NRF_TIMER_MODE_TIMER;
    err = nrfx_timer_init(&m_timer_read, &timer_cfg, timer_handler_read);
    if (err != NRFX_SUCCESS) {
		printk("nrfx_timer_init failed with: %d\n", err);
	}


			
	nrfx_timer_extended_compare(&m_timer_read,
		NRF_TIMER_CC_CHANNEL0,
		nrfx_timer_ms_to_ticks(&m_timer_read, COUNT_READ_INTERVAL),
		NRF_TIMER_SHORT_COMPARE0_CLEAR_MASK,
		true);

	//edited
	nrfx_timer_extended_compare(&m_timer_read,
		NRF_TIMER_CC_CHANNEL1,
		nrfx_timer_ms_to_ticks(&m_timer_read, COUNT_READ_INTERVAL),
		NRF_TIMER_SHORT_COMPARE1_CLEAR_MASK,
		true);
	//edited
	
	IRQ_CONNECT(TIMER0_IRQn, IRQ_PRIO_LOWEST,
		nrfx_timer_0_irq_handler, NULL, 0);
	IRQ_CONNECT(TIMER1_IRQn, IRQ_PRIO_LOWEST,
		nrfx_timer_1_irq_handler, NULL, 0);
	
	nrfx_timer_clear(&m_timer_count);
	nrfx_timer_clear(&m_timer_read);
	
	nrfx_timer_enable(&m_timer_count);
	nrfx_timer_enable(&m_timer_read);
		
}

void ppi_init()
{
    nrfx_err_t err;
	
	/* Allocate a (D)PPI channel. */
#if defined(DPPI_PRESENT)
	uint8_t ppi_channel_1;
	uint8_t ppi_channel_2;
	
	err = nrfx_dppi_channel_alloc(&ppi_channel_1);
	err = nrfx_dppi_channel_alloc(&ppi_channel_2);

	//edited
	uint8_t ppi_channel_3;
	uint8_t ppi_channel_4;
	
	err = nrfx_dppi_channel_alloc(&ppi_channel_3);
	err = nrfx_dppi_channel_alloc(&ppi_channel_4);
	//edited
#else
	nrf_ppi_channel_t ppi_channel_1;
    nrf_ppi_channel_t ppi_channel_2;
	
	err = nrfx_ppi_channel_alloc(&ppi_channel_1);
	err = nrfx_ppi_channel_alloc(&ppi_channel_2);
#endif
	if (err != NRFX_SUCCESS) {
		printk("(D)PPI channel allocation error: %08x", err);
		return;
	}

	
	nrfx_gppi_channel_endpoints_setup(ppi_channel_1,
		nrf_gpiote_event_address_get(NRF_GPIOTE,
			nrfx_gpiote_in_event_get(&gpiote,INPUT_PIN)),
		nrf_timer_task_address_get(m_timer_count.p_reg,
			NRF_TIMER_TASK_COUNT));
			

	nrfx_gppi_channel_endpoints_setup(ppi_channel_2,
		nrf_timer_event_address_get(m_timer_read.p_reg,
			NRF_TIMER_EVENT_COMPARE0),
        nrf_timer_task_address_get(m_timer_count.p_reg, 
			NRF_TIMER_TASK_CAPTURE0));
	
	
    nrfx_gppi_fork_endpoint_setup(ppi_channel_2, nrf_timer_task_address_get(m_timer_count.p_reg,NRF_TIMER_TASK_CLEAR)); // Clear counter timer using (D)PPI.
 

	//edited 
		nrfx_gppi_channel_endpoints_setup(ppi_channel_3,
		nrf_gpiote_event_address_get(NRF_GPIOTE,
			nrfx_gpiote_in_event_get(&gpiote,INPUT_PIN_1)),
		nrf_timer_task_address_get(m_timer_count.p_reg,
			NRF_TIMER_TASK_COUNT));
			

	nrfx_gppi_channel_endpoints_setup(ppi_channel_4,
		nrf_timer_event_address_get(m_timer_read.p_reg,
			NRF_TIMER_EVENT_COMPARE1),
        nrf_timer_task_address_get(m_timer_count.p_reg, 
			NRF_TIMER_TASK_CAPTURE1));
	
	
    nrfx_gppi_fork_endpoint_setup(ppi_channel_4, nrf_timer_task_address_get(m_timer_count.p_reg,NRF_TIMER_TASK_CLEAR)); // Clear counter timer using (D)PPI.
	//edited

	/* Enable (D)PPI channel. */
#if defined(DPPI_PRESENT)
	err = nrfx_dppi_channel_enable(ppi_channel_1);
	err = nrfx_dppi_channel_enable(ppi_channel_2);

	//edited
	err = nrfx_dppi_channel_enable(ppi_channel_3);
	err = nrfx_dppi_channel_enable(ppi_channel_4);
	//edited
#else
	err = nrfx_ppi_channel_enable(ppi_channel_1);
	err = nrfx_ppi_channel_enable(ppi_channel_2);
#endif
	if (err != NRFX_SUCCESS) {
		printk("Failed to enable (D)PPI channel, error: %08x", err);
		return;
	}
}


// Seq value for 500kHz out
//int16_t buf[] = {(1 << 15) |  12  }; // Inverse polarity (bit 15)

// Seq value for 1MHz out
int16_t buf[] = {(1 << 15) |  22  }; // Inverse polarity (bit 15)

#define PWM_PIN (27UL)
static void bare_metal_pwm_test_start(void)
{
	
  // Start accurate HFCLK (XOSC)
  NRF_CLOCK->TASKS_HFCLKSTART = 1;
  while (NRF_CLOCK->EVENTS_HFCLKSTARTED == 0) ;
  NRF_CLOCK->EVENTS_HFCLKSTARTED = 0;
  
  // Configure PWM_PIN as output, and set it to 0
  NRF_P0->DIRSET = (1 << PWM_PIN);
  NRF_P0->OUTCLR = (1 << PWM_PIN);
  
  
  NRF_PWM0->PRESCALER   = PWM_PRESCALER_PRESCALER_DIV_1; // 1 us
  NRF_PWM0->PSEL.OUT[0] = PWM_PIN;
  NRF_PWM0->MODE        = (PWM_MODE_UPDOWN_Up << PWM_MODE_UPDOWN_Pos);
  NRF_PWM0->DECODER     = (PWM_DECODER_LOAD_Common       << PWM_DECODER_LOAD_Pos) | 
                          (PWM_DECODER_MODE_RefreshCount << PWM_DECODER_MODE_Pos);
  NRF_PWM0->LOOP        = (PWM_LOOP_CNT_Disabled << PWM_LOOP_CNT_Pos);
  
  //Countertop for 500kHz out
  //NRF_PWM0->COUNTERTOP = 32; 

  
  //Countertop for 1 Mhz out
  NRF_PWM0->COUNTERTOP = 68;
  
  
  NRF_PWM0->SEQ[0].CNT = ((sizeof(buf) / sizeof(uint16_t)) << PWM_SEQ_CNT_CNT_Pos);
  NRF_PWM0->SEQ[0].ENDDELAY = 0;
  NRF_PWM0->SEQ[0].PTR = (uint32_t)&buf[0];
  NRF_PWM0->SEQ[0].REFRESH = 0;
  NRF_PWM0->SHORTS = 0;
  
  NRF_PWM0->ENABLE = 1;
  NRF_PWM0->TASKS_SEQSTART[0] = 1;
}



void main(void)
{
	gpiote_init();
	
    timer_init();
	
    ppi_init();
    
    printk("Example start\r\n");

    //Start PWM output on PWM_PIN for testing. Connect pin 27 and 28 to test.
    bare_metal_pwm_test_start();

}